"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const ast = require("../parser/ast");
const errors_1 = require("./errors");
const fs_1 = require("fs");
class StdErrorOutput {
    handle(e) {
        if (e instanceof errors_1.TypeError) {
            console.log((e.location.file + " (" + e.location.start.line + "," + e.location.start.column + "): ").yellow + e.message.red);
        }
        else if (e instanceof errors_1.SyntaxError) {
            console.log((ast.currentFile() + " (" + e.location.start.line + "," + e.location.start.column + "): ").yellow + e.message.red);
        }
        else if (e instanceof errors_1.ImportError) {
            if (e.location) {
                console.log((e.location.file + " (" + e.location.start.line + "," + e.location.start.column + "): ").yellow + e.message.red);
            }
            else {
                console.log((e.path + ": ".yellow) + e.message.red);
            }
        }
        else {
            if (e.stack) {
                console.error(this.buildOffendingLineString(e.stack));
                console.error(e.stack);
            }
            else {
                console.error(e);
            }
        }
        process.exit(1);
    }
    /**
     * This generates a string of the offending line in the file in the style of NodeJS.
     * The offending file and line are extracted from the given stack trace.
     * There is definitely room for optimization as it currently only loops through all lines in the file.
     *
     * @param stack The stack trace string.
     */
    buildOffendingLineString(stack) {
        let fileAndLocation = stack
            .split('\n')[1]
            .replace(/\s*at\s*[\w\.]*\s*\(/i, '') // remove the cruft before the path
            .replace(')', '')
            .split(':');
        let file = fileAndLocation[0];
        let lineNumber = Number(fileAndLocation[1]);
        let column = Number(fileAndLocation[2]);
        let outputLine = file + ':' + lineNumber + '\n';
        let currentLineNumber = 1;
        fs_1.readFileSync(file).toString().split('\n').forEach((line) => {
            if (currentLineNumber == lineNumber) {
                outputLine = outputLine + line + '\n';
            }
            currentLineNumber++;
        });
        for (let i = 1; i < column; i++) {
            outputLine = outputLine + ' ';
        }
        return outputLine + '^\n';
    }
}
exports.StdErrorOutput = StdErrorOutput;
//# sourceMappingURL=errorhandler.js.map