"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const types_1 = require("../types");
/**
 * The types defined here must be registered in TypeChecker.checkModule.
 */
class Static {
    static init() {
        Static.t_bool = new types_1.BasicType("bool");
        Static.t_float = new types_1.BasicType("float");
        Static.t_double = new types_1.BasicType("double");
        Static.t_null = new types_1.BasicType("null");
        Static.t_int8 = new types_1.BasicType("int8");
        Static.t_char = new types_1.BasicType("char");
        Static.t_int16 = new types_1.BasicType("int16");
        Static.t_int32 = new types_1.BasicType("int32");
        Static.t_int64 = new types_1.BasicType("int64");
        Static.t_int = new types_1.BasicType("int");
        Static.t_uint8 = new types_1.BasicType("uint8");
        Static.t_byte = new types_1.BasicType("byte");
        Static.t_uint16 = new types_1.BasicType("uint16");
        Static.t_uint32 = new types_1.BasicType("uint32");
        Static.t_uint64 = new types_1.BasicType("uint64");
        // TODO: Depends on arch
        Static.t_uint = new types_1.BasicType("uint");
        Static.t_any = new types_1.BasicType("any");
        Static.t_string = new types_1.BasicType("string");
        Static.t_void = new types_1.BasicType("void");
        Static.t_rune = new types_1.BasicType("rune");
        Static.t_error = new types_1.InterfaceType();
        Static.t_error.name = "error";
        Static.t_voidPtr = new types_1.UnsafePointerType(this.t_void);
        let toError = new types_1.FunctionType();
        toError.name = "toError";
        toError.returnType = Static.t_string;
        toError.objectType = new types_1.RestrictedType(Static.t_error, { isConst: true });
        Static.t_error.methods.set("toError", toError);
        Static.t_coroutine = new types_1.PointerType(Static.t_void, "reference");
    }
    static isInitialized() {
        return Static.t_bool ? true : false;
    }
    static initIfRequired() {
        if (!Static.isInitialized()) {
            Static.init();
        }
    }
}
exports.Static = Static;
//# sourceMappingURL=Static.js.map