"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var GroupKind;
(function (GroupKind) {
    GroupKind[GroupKind["Free"] = 0] = "Free";
    GroupKind[GroupKind["Bound"] = 1] = "Bound";
})(GroupKind = exports.GroupKind || (exports.GroupKind = {}));
class Group {
    constructor(kind, name) {
        this.counter = Group.groupCounter++;
        this.kind = kind;
        if (name) {
            this.name = name;
        }
        else {
            this.name = "$unnamed" + Group.counter++;
        }
    }
    preJoin(scope, loc, doThrow) {
        return this;
    }
    static isLess(g1, g2) {
        if (g1.kind < g2.kind) {
            return true;
        }
        if (g1.kind == g2.kind && g1.counter < g2.counter) {
            return true;
        }
        return false;
    }
    isBound(scope) {
        let g = scope.resolveCanonicalGroup(this);
        return g.kind == GroupKind.Bound;
    }
}
Group.counter = 1;
Group.groupCounter = 0;
exports.Group = Group;
class TupleGroup extends Group {
    constructor(kind, name) {
        super(kind, name);
        this.groups = [];
    }
    preJoin(scope, loc, doThrow) {
        let g = null;
        for (let tg of this.groups) {
            g = g ? scope.joinGroups(g, tg, loc, doThrow) : tg;
        }
        return g;
    }
}
exports.TupleGroup = TupleGroup;
class Taint {
    constructor(group, loc) {
        this.loc = loc;
        this.group = group;
    }
}
exports.Taint = Taint;
function combineRestrictions(r1, r2) {
    if (!r1) {
        return r2;
    }
    if (!r2) {
        return r1;
    }
    return {
        isConst: r1.isConst || r2.isConst
    };
}
exports.combineRestrictions = combineRestrictions;
var GroupCheckFlags;
(function (GroupCheckFlags) {
    GroupCheckFlags[GroupCheckFlags["None"] = 0] = "None";
    GroupCheckFlags[GroupCheckFlags["AllowIsolates"] = 1] = "AllowIsolates";
    GroupCheckFlags[GroupCheckFlags["ForbidIsolates"] = 2] = "ForbidIsolates";
    GroupCheckFlags[GroupCheckFlags["IsolatesMask"] = 3] = "IsolatesMask";
    GroupCheckFlags[GroupCheckFlags["NotIsolateMask"] = 252] = "NotIsolateMask";
    GroupCheckFlags[GroupCheckFlags["NoSideEffects"] = 4] = "NoSideEffects";
    GroupCheckFlags[GroupCheckFlags["AllowUnavailableVariable"] = 8] = "AllowUnavailableVariable";
    GroupCheckFlags[GroupCheckFlags["AllowUnavailableVariableMask"] = 247] = "AllowUnavailableVariableMask";
})(GroupCheckFlags = exports.GroupCheckFlags || (exports.GroupCheckFlags = {}));
//# sourceMappingURL=group.js.map