"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const types_1 = require("../types/");
const scope_1 = require("../scope");
const errors_1 = require("../errors");
const _1 = require("./");
function stripType(t) {
    if (t instanceof types_1.RestrictedType) {
        t = t.elementType;
    }
    return t;
}
exports.stripType = stripType;
function isString(t) {
    if (t instanceof types_1.RestrictedType) {
        return t.elementType == _1.Static.t_string;
    }
    return t == _1.Static.t_string;
}
exports.isString = isString;
function isTupleType(t) {
    if (t instanceof types_1.RestrictedType) {
        return t.elementType instanceof types_1.OrType;
    }
    return t instanceof types_1.TupleType;
}
exports.isTupleType = isTupleType;
function isStringLiteralType(t) {
    if (t instanceof types_1.RestrictedType) {
        return t.elementType instanceof types_1.OrType;
    }
    return t instanceof types_1.StringLiteralType;
}
exports.isStringLiteralType = isStringLiteralType;
function isAny(t) {
    if (t instanceof types_1.RestrictedType) {
        return t.elementType instanceof types_1.OrType;
    }
    return t == _1.Static.t_any;
}
exports.isAny = isAny;
function isOrType(t) {
    if (t instanceof types_1.RestrictedType) {
        return t.elementType instanceof types_1.OrType;
    }
    return t instanceof types_1.OrType;
}
exports.isOrType = isOrType;
function isComplexOrType(t) {
    t = stripType(t);
    if (!(t instanceof types_1.OrType)) {
        return false;
    }
    return !t.stringsOnly();
}
exports.isComplexOrType = isComplexOrType;
function isStringOrType(t) {
    t = stripType(t);
    if (!(t instanceof types_1.OrType)) {
        return false;
    }
    return t.stringsOnly();
}
exports.isStringOrType = isStringOrType;
function isInterface(t) {
    t = stripType(t);
    if (!(t instanceof types_1.PointerType)) {
        return false;
    }
    t = stripType(t.elementType);
    return t instanceof types_1.InterfaceType;
}
exports.isInterface = isInterface;
function isMap(t) {
    t = stripType(t);
    if (!(t instanceof types_1.PointerType)) {
        return false;
    }
    t = stripType(t.elementType);
    return t instanceof types_1.MapType;
}
exports.isMap = isMap;
function isSlice(t) {
    t = stripType(t);
    return t instanceof types_1.SliceType;
}
exports.isSlice = isSlice;
function isArray(t) {
    t = stripType(t);
    return t instanceof types_1.ArrayType;
}
exports.isArray = isArray;
function isUnsafePointer(t) {
    t = stripType(t);
    return t instanceof types_1.UnsafePointerType;
}
exports.isUnsafePointer = isUnsafePointer;
function isNumber(t) {
    t = stripType(t);
    return (t == _1.Static.t_rune || t == _1.Static.t_float || t == _1.Static.t_double || t == _1.Static.t_int || t == _1.Static.t_uint || t == _1.Static.t_byte || t == _1.Static.t_char || t == _1.Static.t_int8 || t == _1.Static.t_int16 || t == _1.Static.t_int32 || t == _1.Static.t_int64 || t == _1.Static.t_uint8 || t == _1.Static.t_uint16 || t == _1.Static.t_uint32 || t == _1.Static.t_uint64);
}
exports.isNumber = isNumber;
function isStruct(t) {
    t = stripType(t);
    return t instanceof types_1.StructType;
}
exports.isStruct = isStruct;
function isTuple(t) {
    t = stripType(t);
    return t instanceof types_1.TupleType;
}
exports.isTuple = isTuple;
function isTemplateType(t) {
    t = stripType(t);
    return t instanceof types_1.TemplateType;
}
exports.isTemplateType = isTemplateType;
function isMutableValue(t) {
    t = stripType(t);
    return t instanceof types_1.StructType || t instanceof types_1.TupleType || t instanceof types_1.ArrayType;
}
exports.isMutableValue = isMutableValue;
function isStrong(t) {
    if (t instanceof types_1.RestrictedType) {
        t = t.elementType;
    }
    if (t instanceof types_1.PointerType && t.mode == "strong") {
        return true;
    }
    if (t instanceof types_1.SliceType && t.mode == "strong") {
        return true;
    }
    return false;
}
exports.isStrong = isStrong;
function isUnique(t) {
    if (t instanceof types_1.RestrictedType) {
        t = t.elementType;
    }
    if (t instanceof types_1.PointerType && t.mode == "unique") {
        return true;
    }
    if (t instanceof types_1.SliceType && t.mode == "unique") {
        return true;
    }
    return false;
}
exports.isUnique = isUnique;
function isReference(t) {
    if (t instanceof types_1.RestrictedType) {
        t = t.elementType;
    }
    if (t instanceof types_1.PointerType && t.mode == "reference") {
        return true;
    }
    if (t instanceof types_1.SliceType && t.mode == "reference") {
        return true;
    }
    return false;
}
exports.isReference = isReference;
function isLocalReference(t) {
    if (t instanceof types_1.RestrictedType) {
        t = t.elementType;
    }
    if (t instanceof types_1.PointerType && t.mode == "local_reference") {
        return true;
    }
    if (t instanceof types_1.SliceType && t.mode == "local_reference") {
        return true;
    }
    return false;
}
exports.isLocalReference = isLocalReference;
function isConst(t) {
    if (t instanceof types_1.RestrictedType) {
        if (t.isConst) {
            return true;
        }
        t = t.elementType;
    }
    return false;
}
exports.isConst = isConst;
function isPlatformIntNumber(type) {
    type = stripType(type);
    return (type == _1.Static.t_int || type == _1.Static.t_uint);
}
exports.isPlatformIntNumber = isPlatformIntNumber;
function isIntNumber(type) {
    type = stripType(type);
    if (type == _1.Static.t_int || type == _1.Static.t_uint || type == _1.Static.t_byte || type == _1.Static.t_char || type == _1.Static.t_int8 || type == _1.Static.t_int16 || type == _1.Static.t_int32 || type == _1.Static.t_int64 || type == _1.Static.t_uint8 || type == _1.Static.t_uint16 || type == _1.Static.t_uint32 || type == _1.Static.t_uint64) {
        return true;
    }
    return false;
}
exports.isIntNumber = isIntNumber;
// TODO: Platform specific
function isInt32Number(t) {
    t = stripType(t);
    return t == _1.Static.t_int32 || t == _1.Static.t_uint32;
}
exports.isInt32Number = isInt32Number;
// TODO: Platform specific
function isUInt32Number(t) {
    t = stripType(t);
    return t == _1.Static.t_uint32;
}
exports.isUInt32Number = isUInt32Number;
function isPrimitive(t) {
    t = stripType(t);
    return (t == _1.Static.t_rune || t == _1.Static.t_bool || t == _1.Static.t_float || t == _1.Static.t_double || t == _1.Static.t_int8 || t == _1.Static.t_int16 || t == _1.Static.t_int32 || t == _1.Static.t_int64 || t == _1.Static.t_uint8 || t == _1.Static.t_uint16 || t == _1.Static.t_uint32 || t == _1.Static.t_uint64 || t == _1.Static.t_null || t == _1.Static.t_void);
}
exports.isPrimitive = isPrimitive;
function isSafePointer(t) {
    t = stripType(t);
    return (t instanceof types_1.PointerType);
}
exports.isSafePointer = isSafePointer;
/**
 * Returns true if the expression denoted by 'node' can be the left-hand-side of an assignment
 */
function isLeftHandSide(node, scope, _allowConstVariable = false) {
    if (node.op == "id") {
        if (!_allowConstVariable) {
            let e = scope.resolveElement(node.value);
            if (!e) {
                throw new errors_1.ImplementationError();
            }
            if ((e instanceof scope_1.Variable && e.isConst) || (e instanceof scope_1.FunctionParameter && e.isConst)) {
                return false;
            }
        }
        return true;
    }
    else if (node.op == "unary*") {
        return true;
    }
    else if (node.op == ".") {
        if (node.lhs.type instanceof types_1.PointerType || node.lhs.type instanceof types_1.UnsafePointerType) {
            return true;
        }
        return isLeftHandSide(node.lhs, scope /*, true*/);
    }
    else if (node.op == "[") {
        if (node.lhs.type instanceof types_1.UnsafePointerType || node.lhs.type instanceof types_1.SliceType) {
            return true;
        }
        if (node.lhs.type == _1.Static.t_string) {
            return false;
        }
        return isLeftHandSide(node.lhs, scope /*, true*/);
    }
    return false;
}
exports.isLeftHandSide = isLeftHandSide;
/**
 * Returns true if the expression denoted by 'node' can be the left-hand-side of an assignment
 */
function isMutable(node, scope) {
    if (node.op == "id") {
        let e = scope.resolveElement(node.value);
        if (!e) {
            throw new errors_1.ImplementationError();
        }
        if ((e instanceof scope_1.Variable && e.isConst) || (e instanceof scope_1.FunctionParameter && e.isConst)) {
            return false;
        }
        return true;
    }
    else if (node.op == "unary*") {
        if (isConst(node.rhs.type)) {
            return false;
        }
        return true;
    }
    else if (node.op == ".") {
        if (isConst(node.lhs.type)) {
            return false;
        }
        if (node.lhs.type instanceof types_1.PointerType || node.lhs.type instanceof types_1.UnsafePointerType) {
            return true;
        }
        return isMutable(node.lhs, scope);
    }
    else if (node.op == "[") {
        if (isConst(node.lhs.type)) {
            return false;
        }
        if (node.lhs.type instanceof types_1.UnsafePointerType || node.lhs.type instanceof types_1.SliceType) {
            return true;
        }
        if (node.lhs.type == _1.Static.t_string) {
            return false;
        }
        return isMutable(node.lhs, scope);
    }
    return !isConst(node.type);
}
exports.isMutable = isMutable;
function isAssignable(node, scope) {
    if (node.op == "id") {
        let e = scope.resolveElement(node.value);
        if (!e) {
            throw new errors_1.ImplementationError();
        }
        if ((e instanceof scope_1.Variable && e.isConst) || (e instanceof scope_1.FunctionParameter && e.isConst)) {
            return false;
        }
        return true;
    }
    else if (node.op == "unary*") {
        if (isConst(node.rhs.type)) {
            return false;
        }
        return true;
    }
    else if (node.op == ".") {
        if (node.lhs.type instanceof types_1.PointerType || node.lhs.type instanceof types_1.UnsafePointerType) {
            if (isConst(node.lhs.type)) {
                return false;
            }
            return true;
        }
        return isAssignable(node.lhs, scope);
    }
    else if (node.op == "[") {
        if (node.lhs.type instanceof types_1.UnsafePointerType || node.lhs.type instanceof types_1.SliceType) {
            if (isConst(node.lhs.type)) {
                return false;
            }
            return true;
        }
        if (node.lhs.type == _1.Static.t_string) {
            return false;
        }
        return isAssignable(node.lhs, scope);
    }
    return false;
}
exports.isAssignable = isAssignable;
/**
 * A pure value contains no pointers and can be copied byte by byte.
 */
function isPureValue(t) {
    t = types_1.RestrictedType.strip(t);
    if (t == _1.Static.t_int || t == _1.Static.t_uint || t == _1.Static.t_byte || t == _1.Static.t_char || t == _1.Static.t_rune || t == _1.Static.t_bool || t == _1.Static.t_float || t == _1.Static.t_double || t == _1.Static.t_int8 || t == _1.Static.t_int16 || t == _1.Static.t_int32 || t == _1.Static.t_int64 || t == _1.Static.t_uint8 || t == _1.Static.t_uint16 || t == _1.Static.t_uint32 || t == _1.Static.t_uint64 || t == _1.Static.t_null || t == _1.Static.t_void) {
        return true;
    }
    if (t instanceof types_1.TupleType) {
        for (let p of t.types) {
            if (!isPureValue(p)) {
                return false;
            }
        }
        return true;
    }
    else if (t instanceof types_1.ArrayType) {
        return isPureValue(t.elementType);
    }
    else if (t instanceof types_1.StructType) {
        for (let f of t.fields) {
            if (!isPureValue(f.type)) {
                return false;
            }
        }
        if (t.extends && !isPureValue(t.extends)) {
            return false;
        }
        return true;
    }
    else if (t instanceof types_1.OrType) {
        return t.isPureValue();
    }
    return false;
}
exports.isPureValue = isPureValue;
function applyConst(t, loc) {
    if (isSafePointer(t)) {
        //            let r: RestrictedType;
        if (t instanceof types_1.RestrictedType) {
            //                r = t;
            t = t.elementType;
        }
        return new types_1.PointerType(makeConst(t.elementType, loc), t.mode);
        //            if (r) {
        //                t =  new RestrictedType(t, r);
        //            }
    }
    else if (isSlice(t)) {
        //            let r: RestrictedType;
        if (t instanceof types_1.RestrictedType) {
            //                r = t;
            t = t.elementType;
        }
        return new types_1.SliceType(makeConst(t.arrayType, loc), t.mode);
        //            if (r) {
        //                t =  new RestrictedType(t, r);
        //            }
    }
    return makeConst(t, loc);
}
exports.applyConst = applyConst;
/**
 * Returns true if the expression yields ownership of the object it is pointing to.
 * Call the function only on expressions of pointer type or expressions that can be assigned to a pointer type
 */
function isTakeExpression(enode) {
    if (enode.op == "clone" || enode.op == "take" || enode.op == "pop" || enode.op == "array" || enode.op == "object" || enode.op == "tuple" || enode.op == "make" || enode.op == "null" || (enode.op == ":" && (isStrong(enode.type) || isUnique(enode.type)))) {
        return true;
    }
    if (enode.op == "(") {
        // If the function returns a reference pointer, it is not a take expression, since it does not yield ownership.
        // In this case it just yields a reference.
        if (isSlice(enode.type) || isSafePointer(enode.type)) {
            return isStrong(enode.type) || isUnique(enode.type);
        }
        return true;
    }
    // A slice operation on a string creates a new string which already has a reference count of 1.
    // Hence it behaves like a take expression.
    // Adding a string, or casting a slice to a string creates a new string, too. Hence, it behaves like a take expression
    if (enode.type == _1.Static.t_string && (enode.op == ":" || enode.op == "+" || (enode.op == "typeCast" && isSlice(enode.rhs.type)))) {
        return true;
    }
    // Casting a string to a slice returns a new slice. Hence, it behaves like a take expression
    if (isSlice(enode.type) && enode.op == "typeCast" && enode.rhs.type == _1.Static.t_string) {
        return true;
    }
    // An expression of the kind &StructType{...} is a take expression
    if (enode.op == "unary&" && enode.rhs.isLiteral()) {
        return true;
    }
    return false;
}
exports.isTakeExpression = isTakeExpression;
/**
 * Returns a Node with op == 'id' if enode is a local variable, or a (series of) typecasts of a local variable.
 * If enode is a part of a local variable (e.g. an array element and the array is a local variable),
 * the function returns null. It only returns true if enode represents the entire(!) value stored in a local variable.
 */
function getUnderlyingLocalVariable(enode) {
    while (enode.op == "typeCast" && !isTakeExpression(enode)) {
        enode = enode.rhs;
    }
    if (enode.op == "id") {
        return enode;
    }
    return null;
}
exports.getUnderlyingLocalVariable = getUnderlyingLocalVariable;
function makeConst(t, loc) {
    if (t instanceof types_1.RestrictedType) {
        if (t.isConst) {
            return t;
        }
        return new types_1.RestrictedType(t.elementType, { isConst: true });
    }
    //        if (Helper.isPrimitive(t)) {
    //            return t;
    //        }
    return new types_1.RestrictedType(t, { isConst: true });
}
exports.makeConst = makeConst;
function isSigned(t) {
    t = stripType(t);
    if (t == _1.Static.t_char || t == _1.Static.t_int || t == _1.Static.t_int8 || t == _1.Static.t_int16 || t == _1.Static.t_int32 || t == _1.Static.t_int64 || t == _1.Static.t_float || t == _1.Static.t_double) {
        return true;
    }
    if (t == _1.Static.t_byte || t == _1.Static.t_uint || t == _1.Static.t_uint8 || t == _1.Static.t_uint16 || t == _1.Static.t_uint32 || t == _1.Static.t_uint64) {
        return false;
    }
    if (isUnsafePointer(t)) {
        return true;
    }
    throw new errors_1.ImplementationError("CodeGen: Implementation error: signed check on non number type " + t.toString());
}
exports.isSigned = isSigned;
function isPureLiteral(t, n) {
    t = types_1.RestrictedType.strip(t);
    if (t instanceof types_1.InterfaceType) {
        return false;
    }
    switch (n.op) {
        case "int":
        case "bool":
        case "float":
        case "null":
        case "rune":
        case "str":
            return true;
        case "array":
            {
                if (isArray(t)) {
                    if (n.parameters) {
                        for (let p of n.parameters) {
                            if (p.op == "unary...") {
                                continue;
                            }
                            else if (!this.isPureLiteral(t.elementType, p)) {
                                return false;
                            }
                        }
                    }
                    return true;
                }
                break;
            }
        case "tuple":
            {
                let i = 0;
                for (let p of n.parameters) {
                    if (!this.isPureLiteral(t.types[i], p)) {
                        return false;
                    }
                }
                return true;
            }
        case "object":
            {
                if (isStruct(t)) {
                    if (n.parameters) {
                        for (let p of n.parameters) {
                            let f = t.field(p.name.value);
                            if (!this.isPureLiteral(f.type, p.lhs)) {
                                return false;
                            }
                        }
                    }
                    return true;
                }
                break;
            }
    }
    return false;
}
exports.isPureLiteral = isPureLiteral;
class BinaryBuffer {
    constructor(size = 1024) {
        this.offset = 0;
        this.len = 0;
        this.numberArrayBuf = new ArrayBuffer(8);
        this.arrayBuf = new ArrayBuffer(size);
        this.buf = new Uint8Array(this.arrayBuf);
        this.numberBuf = new Uint8Array(this.numberArrayBuf);
        this.numberUint8 = new Uint8Array(this.numberArrayBuf);
        this.numberUint16 = new Uint16Array(this.numberArrayBuf);
        this.numberUint32 = new Uint32Array(this.numberArrayBuf);
        this.numberInt8 = new Int8Array(this.numberArrayBuf);
        this.numberInt16 = new Int16Array(this.numberArrayBuf);
        this.numberInt32 = new Int32Array(this.numberArrayBuf);
        this.numberFloat32 = new Float32Array(this.numberArrayBuf);
        this.numberFloat64 = new Float64Array(this.numberArrayBuf);
    }
    get data() {
        return new Uint8Array(this.arrayBuf, 0, this.len);
    }
    get arrayBuffer() {
        return this.arrayBuf;
    }
    get length() {
        return this.len;
    }
    fill(length) {
        if (this.buf.byteLength < this.offset + length) {
            this.resize(Math.max(this.buf.byteLength * 2, this.buf.byteLength + length));
        }
        for (let i = 0; i < length; i++) {
            this.buf[this.offset + i] = 0;
        }
        this.offset += length;
        this.len = Math.max(this.len, this.offset);
    }
    appendUint8(n) {
        this.buf[this.offset] = n;
        this.offset++;
        this.len = Math.max(this.len, this.offset);
    }
    appendInt8(n) {
        this.numberInt8[0] = n;
        this.buf[this.offset] = this.numberBuf[0];
        this.offset++;
        this.len = Math.max(this.len, this.offset);
    }
    appendUint16(n) {
        this.numberUint16[0] = n;
        this.buf[this.offset] = this.numberBuf[0];
        this.buf[this.offset + 1] = this.numberBuf[1];
        this.offset += 2;
        this.len = Math.max(this.len, this.offset);
    }
    appendInt16(n) {
        this.numberInt16[0] = n;
        this.buf[this.offset] = this.numberBuf[0];
        this.buf[this.offset + 1] = this.numberBuf[1];
        this.offset += 2;
        this.len = Math.max(this.len, this.offset);
    }
    appendUint32(n) {
        this.numberUint32[0] = n;
        this.buf[this.offset] = this.numberBuf[0];
        this.buf[this.offset + 1] = this.numberBuf[1];
        this.buf[this.offset + 2] = this.numberBuf[2];
        this.buf[this.offset + 3] = this.numberBuf[3];
        this.offset += 4;
        this.len = Math.max(this.len, this.offset);
    }
    appendInt32(n) {
        this.numberInt32[0] = n;
        this.buf[this.offset] = this.numberBuf[0];
        this.buf[this.offset + 1] = this.numberBuf[1];
        this.buf[this.offset + 2] = this.numberBuf[2];
        this.buf[this.offset + 3] = this.numberBuf[3];
        this.offset += 4;
        this.len = Math.max(this.len, this.offset);
    }
    // TODO: Big numbers
    appendUint64(n) {
        this.numberUint32[0] = n;
        this.buf[this.offset] = this.numberBuf[0];
        this.buf[this.offset + 1] = this.numberBuf[1];
        this.buf[this.offset + 2] = this.numberBuf[2];
        this.buf[this.offset + 3] = this.numberBuf[3];
        this.buf[this.offset + 4] = 0;
        this.buf[this.offset + 5] = 0;
        this.buf[this.offset + 6] = 0;
        this.buf[this.offset + 7] = 0;
        this.offset += 8;
        this.len = Math.max(this.len, this.offset);
    }
    // TODO: Big numbers
    appendInt64(n) {
        this.numberInt32[0] = n;
        this.buf[this.offset] = this.numberBuf[0];
        this.buf[this.offset + 1] = this.numberBuf[1];
        this.buf[this.offset + 2] = this.numberBuf[2];
        this.buf[this.offset + 3] = this.numberBuf[3];
        this.buf[this.offset + 4] = n < 0 ? 255 : 0;
        this.buf[this.offset + 5] = n < 0 ? 255 : 0;
        this.buf[this.offset + 6] = n < 0 ? 255 : 0;
        this.buf[this.offset + 7] = n < 0 ? 255 : 0;
        this.offset += 8;
        this.len = Math.max(this.len, this.offset);
    }
    appendFloat32(n) {
        this.numberFloat32[0] = n;
        this.buf[this.offset] = this.numberBuf[0];
        this.buf[this.offset + 1] = this.numberBuf[1];
        this.buf[this.offset + 2] = this.numberBuf[2];
        this.buf[this.offset + 3] = this.numberBuf[3];
        this.offset += 4;
        this.len = Math.max(this.len, this.offset);
    }
    appendFloat64(n) {
        this.numberFloat64[0] = n;
        this.buf[this.offset] = this.numberBuf[0];
        this.buf[this.offset + 1] = this.numberBuf[1];
        this.buf[this.offset + 2] = this.numberBuf[2];
        this.buf[this.offset + 3] = this.numberBuf[3];
        this.buf[this.offset + 4] = this.numberBuf[4];
        this.buf[this.offset + 5] = this.numberBuf[5];
        this.buf[this.offset + 6] = this.numberBuf[6];
        this.buf[this.offset + 7] = this.numberBuf[7];
        this.offset += 8;
        this.len = Math.max(this.len, this.offset);
    }
    // TODO: 64bit pointer support
    appendPointer(n) {
        this.appendUint32(n);
    }
    resize(size) {
        this.arrayBuf = transferArrayBuffer(this.buf.buffer, size);
        this.buf = new Uint8Array(this.arrayBuf);
    }
}
exports.BinaryBuffer = BinaryBuffer;
function transferArrayBuffer(source, length) {
    source = Object(source);
    var dest = new ArrayBuffer(length);
    if (!(source instanceof ArrayBuffer) || !(dest instanceof ArrayBuffer)) {
        throw new TypeError('Source and destination must be ArrayBuffer instances');
    }
    if (dest.byteLength >= source.byteLength) {
        var nextOffset = 0;
        var leftBytes = source.byteLength;
        var wordSizes = [8, 4, 2, 1];
        wordSizes.forEach(function (_wordSize_) {
            if (leftBytes >= _wordSize_) {
                var done = transferWith(_wordSize_, source, dest, nextOffset, leftBytes);
                nextOffset = done.nextOffset;
                leftBytes = done.leftBytes;
            }
        });
    }
    return dest;
    function transferWith(wordSize, source, dest, nextOffset, leftBytes) {
        var ViewClass = Uint8Array;
        switch (wordSize) {
            case 8:
                ViewClass = Float64Array;
                break;
            case 4:
                ViewClass = Float32Array;
                break;
            case 2:
                ViewClass = Uint16Array;
                break;
            case 1:
                ViewClass = Uint8Array;
                break;
            default:
                ViewClass = Uint8Array;
                break;
        }
        var view_source = new ViewClass(source, nextOffset, Math.trunc(leftBytes / wordSize));
        var view_dest = new ViewClass(dest, nextOffset, Math.trunc(leftBytes / wordSize));
        for (var i = 0; i < view_dest.length; i++) {
            view_dest[i] = view_source[i];
        }
        return {
            nextOffset: view_source.byteOffset + view_source.byteLength,
            leftBytes: source.byteLength - (view_source.byteOffset + view_source.byteLength)
        };
    }
}
//# sourceMappingURL=helper.js.map