"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const errors_1 = require("../errors");
const Type_1 = require("./Type");
class InterfaceType extends Type_1.Type {
    constructor() {
        super(...arguments);
        this.extendsInterfaces = [];
        /**
         * Member methods indexed by their name
         */
        this.methods = new Map();
        this.sortedMethodNames = [];
        /**
         * Required during recursive checking
         */
        this._markChecked = false;
    }
    getAllMethods(map) {
        if (!map) {
            map = new Map();
        }
        for (let key of this.methods.keys()) {
            map.set(key, this.methods.get(key));
        }
        for (let b of this.extendsInterfaces) {
            if (!(b instanceof InterfaceType)) {
                continue;
            }
            b.getAllMethods(map);
        }
        return map;
    }
    getAllBaseTypes(base) {
        if (base && base.indexOf(this) != -1) {
            return base;
        }
        for (let b of this.extendsInterfaces) {
            if (!(b instanceof InterfaceType)) {
                continue;
            }
            if (!base) {
                base = [b];
            }
            else {
                base.push(b);
            }
            base = b.getAllBaseTypes(base);
        }
        return base;
    }
    hasBaseType(b) {
        for (let i of this.extendsInterfaces) {
            if (i == b) {
                return true;
            }
            if (i.hasBaseType(b)) {
                return true;
            }
        }
        return false;
    }
    toString() {
        if (this.name) {
            return this.name;
        }
        if (this.extendsInterfaces.length > 0 || this.methods.size > 0) {
            return "interface{...}";
        }
        return "interface{}";
    }
    // TODO: Scoping
    // TODO: Unnamned interfaces
    toTypeCodeString() {
        if (this.name) {
            return this.name;
        }
        throw new errors_1.TodoError();
    }
    method(name) {
        if (this.methods.has(name)) {
            return this.methods.get(name);
        }
        for (let iface of this.extendsInterfaces) {
            if (iface instanceof InterfaceType) {
                let m = iface.method(name);
                if (m) {
                    return m;
                }
            }
        }
        return null;
    }
    methodIndex(name) {
        this.sortMethodNames();
        let index = this.sortedMethodNames.indexOf(name);
        if (index == -1) {
            throw new errors_1.ImplementationError(name);
        }
        return index;
    }
    sortMethodNames() {
        if (this.sortedMethodNames.length != 0) {
            return this.sortedMethodNames;
        }
        for (let i = 0; i < this.extendsInterfaces.length; i++) {
            let iface = this.extendsInterfaces[i];
            if (iface instanceof InterfaceType) {
                iface.sortMethodNames();
            }
            this.sortedMethodNames = this.sortedMethodNames.concat(iface.sortedMethodNames);
        }
        if (this.sortedMethodNames.length == 0) {
            this.sortedMethodNames.push("__dtr__");
        }
        let names = [];
        for (let name of this.methods.keys()) {
            names.push(name);
        }
        names.sort();
        this.sortedMethodNames = this.sortedMethodNames.concat(names);
        return this.sortedMethodNames;
    }
}
exports.InterfaceType = InterfaceType;
//# sourceMappingURL=InterfaceType.js.map