"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const Type_1 = require("./Type");
const ArrayType_1 = require("./ArrayType");
const RestrictedType_1 = require("./RestrictedType");
const errors_1 = require("../errors");
class SliceType extends Type_1.Type {
    constructor(arrayType, mode) {
        super();
        if (arrayType instanceof RestrictedType_1.RestrictedType && !(RestrictedType_1.RestrictedType.strip(arrayType) instanceof ArrayType_1.ArrayType)) {
            throw new errors_1.ImplementationError('slice of a non-array');
        }
        this.arrayType = arrayType;
        this.mode = mode;
    }
    array() {
        if (this.arrayType instanceof ArrayType_1.ArrayType) {
            return this.arrayType;
        }
        return this.arrayType.elementType;
    }
    getElementType() {
        if (this.arrayType instanceof ArrayType_1.ArrayType) {
            return this.arrayType.elementType;
        }
        return this.arrayType.elementType.elementType;
    }
    toString() {
        if (this.name) {
            return this.name;
        }
        let mode = "";
        if (this.mode == "local_reference") {
            mode = "&";
        }
        else if (this.mode == "reference") {
            mode = "~";
        }
        else if (this.mode == "unique") {
            mode = "^";
        }
        return mode + "[]" + this.array().elementType.toString();
    }
    toTypeCodeString() {
        return this.mode.toString() + "[]" + this.array().elementType.toString();
    }
}
exports.SliceType = SliceType;
//# sourceMappingURL=SliceType.js.map