"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const Type_1 = require("./Type");
class StructType extends Type_1.Type {
    constructor() {
        super();
        this.implements = [];
        /**
         * Fields of the struct, ordered by their appearance in the code
         */
        this.fields = [];
        /**
         * Member methods indexed by their name
         */
        this.methods = new Map();
    }
    field(name, ownFieldsOnly = false) {
        for (let f of this.fields) {
            if (f.name == name) {
                return f;
            }
        }
        if (!ownFieldsOnly && this.extends && this.extends instanceof StructType) {
            return this.extends.field(name);
        }
        return null;
    }
    method(name) {
        if (this.methods.has(name)) {
            return this.methods.get(name);
        }
        if (this.extends && this.extends instanceof StructType) {
            return this.extends.method(name);
        }
        return null;
    }
    toString() {
        if (this.name) {
            return this.name;
        }
        let str = "struct{";
        str += this.fields.join(",");
        str += "}";
        return str;
    }
    toTypeCodeString() {
        if (this.name) {
            return this.name;
        }
        let str = "struct{";
        str += this.fields.map(function (f) { return f.toTypeCodeString(); }).join(",");
        str += "}";
        return str;
    }
    getAllMethodsAndFields(map) {
        if (!map) {
            map = new Map();
        }
        for (let key of this.methods.keys()) {
            map.set(key, this.methods.get(key));
        }
        for (let f of this.fields) {
            map.set(f.name, f);
        }
        return map;
    }
    getAllBaseTypes(base) {
        if (base && base.indexOf(this) != -1) {
            return base;
        }
        if (this.extends) {
            if (!base) {
                base = [this.extends];
            }
            else {
                base.push(this.extends);
            }
            if (this.extends instanceof StructType) {
                base = this.extends.getAllBaseTypes(base);
            }
        }
        return base;
    }
    doesExtend(parent) {
        if (this.extends == parent) {
            return true;
        }
        else if (this.extends && (this.extends instanceof StructType)) {
            return this.extends.doesExtend(parent);
        }
        return false;
    }
}
exports.StructType = StructType;
/**
 * StructField describes the field of a StructType.
 */
class StructField {
    toString() {
        if (!this.name) {
            return this.type.toString();
        }
        return this.name + " " + this.type.toString();
    }
    toTypeCodeString() {
        if (!this.name) {
            return this.type.toTypeCodeString();
        }
        return this.name;
    }
}
exports.StructField = StructField;
//# sourceMappingURL=StructType.js.map