"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const errors_1 = require("../errors");
const Type_1 = require("./Type");
/**
 * TemplateType can either be a template function or a template struct.
 * Template types have template parameters which are type-wildcards with optional constraints.
 * The template type can be instantiated to become a TemplateFunctionType or a TemplateStructType
 * by binding concrete types to these type-wildcards.
 */
class TemplateType extends Type_1.Type {
    constructor() {
        super();
        this.methods = [];
        this.templateParameterTypes = [];
        this.templateParameterNames = [];
    }
    toTypeCodeString() {
        throw new errors_1.ImplementationError("Typecode of template type is not allowed");
    }
    toString() {
        let g = "<";
        let lst = [];
        for (let i = 0; i < this.templateParameterNames.length; i++) {
            let s = this.templateParameterNames[i];
            if (this.templateParameterTypes[i]) {
                s += " is " + this.templateParameterTypes[i].toString();
            }
            lst.push(s);
        }
        g += lst.join(",");
        g += ">";
        if (this.name) {
            return this.name + g;
        }
        return "template" + g;
    }
}
exports.TemplateType = TemplateType;
//# sourceMappingURL=TemplateType.js.map